
% All programs in this collection are free software: 
% you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright 2007-2008 Alexander Zien.
% Max Planck Intitute for Biological Cybernetics, Tbingen, Germany.
% Friedrich Miescher Lab of the Max Planck Society, Tbingen, Germany.


%==============================================================================
function [ x, res, optParam ] = MP_solve( optParam, MP, x0 );
%==============================================================================

% Generic interface to solvers for mathematical programming.
% Currently supports linear programming (LP) and quadratic
% programming (QP).


%------------------------------------------------------------------------------
% init
%------------------------------------------------------------------------------

typeNum = strmatch( MP.type, {'LP','QP'}, 'exact' );
ASSERT( length(typeNum) == 1 );
res = [];
t0 = cputime;
optParam = setDefault( optParam, 'verbosity', 0 );

% --- license
optParam = setDefault( optParam, 'licence', [] );
if( isempty(optParam.licence) )
  ownLicence = 1;
  [ optParam ] = MP_getLicence( optParam );
else
  ownLicence = 0;
end;


switch( optParam.optimizer )
  
%------------------------------------------------------------------------------
% --- matlab optimization toolbox
%------------------------------------------------------------------------------
 case 'toolbox'
  options = optimset( 'Display', 'off', 'LargeScale', 'off' );
  Aeq = full( MP.A( 1:MP.neq, : ) );
  beq = full( MP.b( 1:MP.neq ) );
  Als = full( MP.A( (MP.neq+1):end, : ) );
  bls = full( MP.b( (MP.neq+1):end ) );
  switch( MP.type )
   case 'LP'
    if( isempty(x0) & 0 )
      [ x, fval, flag, output, lambda ] = linprog( MP.f, Als, bls, Aeq, beq, MP.LB, MP.UB );
    else
      [ x, fval, flag, output, lambda ] = linprog( MP.f, Als, bls, Aeq, beq, MP.LB, MP.UB, x0, options );
    end;
   case 'QP'
    if( isempty(x0) & 0 )
      [ x, fval, flag, output, lambda ] = quadprog( full(MP.Q), MP.c, Als, bls, Aeq, beq, MP.LB, MP.UB );
    else
      [ x, fval, flag, output, lambda ] = quadprog( full(MP.Q), MP.c, Als, bls, Aeq, beq, MP.LB, MP.UB, x0, options );
    end;
  end;
  ASSERT( flag >= 0 );
  res.objVal = fval;
  res.lambda = lambda;
  
%------------------------------------------------------------------------------
% --- CPLEX, using interface by Gunnar Raetsch
%------------------------------------------------------------------------------
 case 'cplexgun'
  A = sparse( MP.A );
  thereWereProblems = 1;
  while( thereWereProblems )
    thereWereProblems = 0;
    try
      switch( MP.type )
       case 'LP'
	[ x, lambda, how, p_lp ] = lp_solve( optParam.licence, MP.f, A, MP.b, MP.LB, MP.UB, MP.neq, optParam.verbosity );
	lp_close( optParam.licence, p_lp );
	fval = x' * MP.f;
       case 'QP'
	if( optParam.verbosity >= 1 )
	  fprintf( '\n' );
	  fprintf( '* solving QP: CPLEX with %s *\n', optParam.cplexQpSolver );
	end;
	ASSERT( issparse(MP.Q) );
	if( isempty(x0) )
	  [ x, lambda, how ] = qp_solve( optParam.licence, MP.Q, MP.c, A, MP.b, MP.LB, MP.UB, MP.neq, optParam.verbosity, optParam.cplexQpSolver );
	else
	  [ x, lambda, how ] = qp_solve_start( optParam.licence, MP.Q, MP.c, A, MP.b, MP.LB, MP.UB, MP.neq, optParam.verbosity, optParam.cplexQpSolver, x0 );
	end;
	fval = x'*MP.c + x'*MP.Q*x;
      end;
    catch
      fprintf( '%s failed\n', MP.type );
      pause( 7 );
      thereWereProblems = 1;
    end;
  end;
  ASSERT( strcmp(how,'OK') );
  res.objVal = fval;
  res.lambda = lambda;
  
%------------------------------------------------------------------------------
% --- CPLEX, using "cplexint" interface (http://control.ee.ethz.ch/~hybrid/cplexint.php)
%------------------------------------------------------------------------------
 case 'cplexint'
  switch( MP.type )
   case 'LP'
    Q = [];
    c = MP.f;
   case 'QP'
    Q = MP.Q;
    c = MP.c;
  end;
  % - use default solver
  qqq = [];
  qqq.int = [1063,0];
  % - hang on to CPLEX license
  options = [];
  options.lic_rel = 10000;
  % - all continuous variables
  varType = repmat( 'C', size(c) );
  [ x, fmin, solstat, details ] = cplexint( Q, c, MP.A, MP.b, 1:MP.neq, [], MP.LB, MP.UB, varType, qqq, options );
  if( solstat ~= 1 )
    fprintf( '%s failed\n', MP.type );
  end;
  res.objVal = fmin;
  res.solstat = solstat;
  res.details = details;
  
%------------------------------------------------------------------------------
% unknown optimizer
%------------------------------------------------------------------------------
 otherwise
  error( 'unknown optimizer' );
end;


%------------------------------------------------------------------------------
% done
%------------------------------------------------------------------------------

if( ownLicence )
  [ optParam ] = MP_releaseLicence( optParam );
end;

res.time = cputime - t0;
if( optParam.verbosity >= 1 )
  fprintf( '%s solved in %.0f secs.\n', optParam.cplexQpSolver, res.time );
  fprintf( '\n' );
end;


