
% All programs in this collection are free software: 
% you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright 2007-2008 Alexander Zien.
% Max Planck Intitute for Biological Cybernetics, Tbingen, Germany.
% Friedrich Miescher Lab of the Max Planck Society, Tbingen, Germany.


%==============================================================================
function [ betas, baseRes, mklRes ] = mklSilp( baseParam, mklParam, Kxs, Y, idxTrain );
%==============================================================================

% Generic SILP wrapper to endow a learning algorithm with
% MKL (multiple kernel learning) capabilities.

% INPUT:
%   Kxs [n*n*P]     "P" kernel matrices for the same "n" data points
%   Y [n*?]         labels of the "n" data points
%   mklParam.baseLearner( baseParam, start, K, Y );       trains base learner (BL), optinal "start" for hotstart
%   mklParam.sumOfAlphas( baseParam, baseRes, Y );        returns sum of dual variables of trained BL
%   mklParam.evalRegularizer( baseParam, baseRes, K );    returns "alpha'*K*alpha" for trained BL and kernel "K"


%==============================================================================
% INIT
%==============================================================================

% === init
nofKernelsX = size( Kxs, 3 );
nofKernelsY = 1;
nofKernels = nofKernelsX * nofKernelsY;
mAll = size( Y, 1 );
if( ~exist('idxTrain','var') )
  idxTrain = (1:mAll)';
end;
mklRes = [];
mTrain = length( idxTrain );

% === check
% --- dims
ASSERT( ndims(Kxs) == 2 | ndims(Kxs) == 3 );
ASSERT( size(Kxs,1) == mAll );
ASSERT( size(Kxs,2) == mAll );
ASSERT( size(idxTrain) == [mTrain,1] );
% --- content
ASSERT( isfinite(Kxs) );
%ASSERT( isfunction( mklParam.baseLearner ) );
%ASSERT( isfunction( mklParam.evalRegularizer ) );
%ASSERT( isfunction( mklParam.sumOfAlphas ) );

% === mklParams
ASSERT( isfield(mklParam,'optParam') );
mklParam = setDefault( mklParam, 'verbosity', 1 );
% --- model
mklParam = setDefault( mklParam, 'normKernels', 'ver' );
mklParam = setDefault( mklParam, 'maxBeta', 1 );
% --- optimization
mklParam = setDefault( mklParam, 'epsRidge', 1e-10 );
mklParam = setDefault( mklParam, 'epsBeta',  1e-12 );
% --- convergence criteria
mklParam = setDefault( mklParam, 'deltaObj', 0 );
mklParam = setDefault( mklParam, 'deltaBetas', 1e-2 );


%==============================================================================
% Prepare SILP
%==============================================================================

optParam = mklParam.optParam;

% === kernels
kronKyKx = cell( nofKernelsX, nofKernelsY );
scales = zeros( nofKernelsX, nofKernelsY );
for( kx = 1:nofKernelsX )
  for( ky = 1:nofKernelsY )
    if( mklParam.verbosity >= 2 )
      fprintf( 'SILP: kernel (%d,%d)\n', kx, ky );
    end;
    K = Kxs( idxTrain, idxTrain, kx );
    switch( mklParam.normKernels )
     case 'none'
      scale = 1;
     case 'var'
      scale = mean(diag(K)) - mean(K(:));
     case 'trace'
      scale = trace( K );
     otherwise
      error( 'unknown kernel normalization' );
    end;
    scales(kx,ky) = scale;
    if( scale ~= 1 )
      K = (1/scale) * K;
    end;
    K = K + mklParam.epsRidge * mean(diag(K)) * speye(mTrain);
    kronKyKx{ kx, ky } = K;
  end;
end;
clear K;
if( mklParam.verbosity >= 2 )
  fprintf( '\n' );
end;
kronKyKx = kronKyKx(:);
ASSERT( size(kronKyKx) == [nofKernels,1] );
mklRes.scales = scales;

% === LP to estimate betas [grows during iterations]
SILP = [];
SILP.type = 'LP';
SILP.verbosity = 0;
% --- objective
SILP.f = zeros( nofKernels+1, 1 );  % betas are not in objective
SILP.f(end) = -1;  % -theta, the last variable, is the objective
% --- constraints 
SILP.A = ones( 1, nofKernels+1 );
SILP.A(end) = 0;
SILP.b = 1;
SILP.neq = 1;
% --- bounds
SILP.LB = zeros( nofKernels+1, 1 );
SILP.LB(end) = -inf;
SILP.UB = mklParam.maxBeta * ones( nofKernels+1, 1 );
SILP.UB(end) = +inf;

% === auxilliary
theta = 0;
betas = ones( nofKernelsX, nofKernelsY ) / nofKernels;
oldTheta = nan;
oldBetas = nan * betas;


%==============================================================================
% Solve SILP
%==============================================================================

% === iterate optimizing alphas / betas
iter = 0;
baseRes = [];
mklRes.thetas = [];
mklRes.betas = {};
mklRes.baseResults = {};
while( 1 )
  iter = iter + 1;
  
  % === base learner
  K = zeros( mTrain, mTrain );
  idxBetas = find( betas ~= 0 );
  for ( k = idxBetas(:)' )
    beta = betas( k );
    ASSERT( beta > 0 );
    K = K + beta * kronKyKx{k};
  end;
  baseRes = feval( mklParam.baseLearner, baseParam, baseRes, K, Y(idxTrain,:) );
  % --- record progress
  mklRes.betas{iter} = betas;
  mklRes.thetas(iter) = theta;
  mklRes.baseResults{iter} = baseRes;
  
  % === convergence?
  theta = baseRes.objVal;
  dTheta = abs( theta - oldTheta );
  dBetas = norm( betas - oldBetas, 'fro' );
  % --- output
  if( mklParam.verbosity >= 1 )
    fprintf( 'SILP: iter %2d: theta=%.4f, d_theta=%.4f, d_B=%.4f \n', iter, theta, dTheta, dBetas );
  end;
  % --- convergence check
  if( nofKernels == 1 )
    break;
  end;
  if( iter > 1 )
    if( dTheta < mklParam.deltaObj )
      break;
    end;
    if( dBetas < mklParam.deltaBetas )
      break;
    end;
  end;
  
  % === LP
  % --- generate new column
  gammas = repmat( nan, nofKernels, 1 );
  for( k = 1:nofKernels )
    gammas(k) = feval( mklParam.evalRegularizer, baseParam, baseRes, kronKyKx{k} );
  end;
  ASSERT( size(gammas) == [nofKernelsX*nofKernelsY,1] );
  sumAlpha = feval( mklParam.sumOfAlphas, baseParam, baseRes, Y(idxTrain,:) );
  SILP.A( end+1, : ) = [ -0.5*gammas' 1 ];
  SILP.b( end+1, 1 ) = sumAlpha;
  % --- start value
  x = [ betas(:) ; 0 ];
  t = SILP.A*x - SILP.b;
  x(end) = -max(t);
  t = SILP.A*x - SILP.b;
  if( max(t) >= 1e-12 )
    fprintf( 'SILP: WARNING: numerical infeasibility = %e\n', max(t) );
  end;
  ASSERT( t <= 1e-6 );
  % --- solve LP -> new betas
  oldTheta = theta;
  oldBetas = betas;
  [ optParam ] = MP_getLicence( optParam );
  [ x, silpRes ] = MP_solve( optParam, SILP, x );
  [ optParam ] = MP_releaseLicence( optParam );
  theta = x(end);
  betas = reshape( x(1:(end-1)), nofKernelsX, nofKernelsY );
  betas( betas < mklParam.epsBeta ) = 0;
  % --- output
  if( mklParam.verbosity >= 2 )
    betaStr = sprintf( '%6.4f ', betas );
    fprintf( 'SILP:  betas = %s\n', betaStr );
  end;
  
end;


%==============================================================================
% DONE
%==============================================================================

% === output
if( mklParam.verbosity == 1 )
  betaStr = sprintf( '%6.4f ', betas );
  fprintf( 'SILP:  betas = %s\n', betaStr );
end;

